"""
    Модуль содержит вспомогательные функции для работы с экшнами
"""


import copy

from google.protobuf.struct_pb2 import Struct
from perxis.items import items_pb2


ACTIONS_COLLECTION_ID = "space_actions"


def process_action_id(action_id: str) -> str:
    """
        Функция для преобразования идентификатора действия чтобы в дальнейшем его можно было
        использовать в Servicer. Примеры преобразований:
        - some-action -> some_action
        - some other action -> some_other_action

        Arguments:
            action_id (str): идентификатор действия

        Returns:
            str: обработанный идентификатор действия
    """
    action_id = action_id.lower().replace("-", "_").replace(" ", "_")

    return action_id


def make_action_dict(
        extension_id: str,
        action_id: str,
        name: str,
        kind: int,
        **kwargs,
) -> dict:
    """
        Функция для создания dict нужной структуры. В дальнейшем он идёт в свойство data item'а

        Arguments:
            extension_id (str): идентификатор расширения
            action_id (str): идентификатор действия
            name (str): название action'а
            kind (str): область действия action'а согласно [документации](https://docs.perxis.ru/docs/api/extension/#actionkind)
            kwargs: любые доп. параметры согласно [документации](https://docs.perxis.ru/docs/api/extension/#action)
        Returns:
            dict
    """
    action_id = process_action_id(action_id)

    return {
        "id": action_id,
        "action": f"grpc:///{extension_id}/{action_id}",
        "name": name,
        "kind": kind,
        **kwargs,
    }


def make_action_struct(data: dict) -> Struct:
    """
        Вспомогательная функция для создания объекта Struct на основании данных из dict

        Arguments:
             data (dict)
        Returns:
             google.protobuf.struct_pb2.Struct
    """
    struct = Struct()
    struct.update(data)

    return struct


def make_action_item(space_id: str, env_id: str, data: dict) -> items_pb2.Item:
    """
        Функция для создания items_pb2.Item с данными action'а

        Arguments:
              space_id (str): идентификатор пространства в котором будет создан action
              env_id (str): идентификатор окружения в котором будет создан action
              data (dict): данные action'а

        Returns:
             items_pb2.Item
    """

    copied_data = copy.deepcopy(data)
    action_id = copied_data.pop("id")

    struct = make_action_struct(data)

    return items_pb2.Item(
        id=action_id,
        space_id=space_id,
        env_id=env_id,
        collection_id=ACTIONS_COLLECTION_ID,
        data=struct
    )
