"""
    Модуль содержит функцию bootstrap для упрощения запуска сервисов расширений
"""


import asyncio

import aiocron
import grpc
import logging

from concurrent import futures
from perxis.collections import collections_pb2_grpc
from perxis.collaborators import collaborators_pb2_grpc
from perxis.invitations import invitations_pb2_grpc
from perxis.environments import environments_pb2_grpc
from perxis.locales import locales_pb2_grpc
from perxis.members import members_pb2_grpc
from perxis.organizations import organizations_pb2_grpc
from perxis.users import users_pb2_grpc
from perxis.references import references_pb2_grpc
from perxis.images import images_pb2_grpc
from perxis.files import files_pb2_grpc
from perxis.roles import roles_pb2_grpc
from perxis.items import items_pb2_grpc
from perxis.spaces import spaces_pb2_grpc
from perxis.clients import clients_pb2_grpc
from perxis.common import operation_service_pb2_grpc
from perxis.extensions import manager_service_pb2_grpc, manager_service_pb2, extension_service_pb2_grpc
from perxis.interceptors import header_adder_interceptor

logger = logging.getLogger(__name__)


async def _main(
        ext_descriptor: manager_service_pb2.ExtensionDescriptor,
        servicer_cls: extension_service_pb2_grpc.ExtensionServiceServicer,
        ext_manager_host: str,
        content_host: str,
        account_host: str,
        files_host: str,
        images_host: str
):
    interceptor = header_adder_interceptor(
        'x-perxis-access', 'system'
    )

    async def _register_extension():
        await ext_manager_stub.RegisterExtensions(manager_service_pb2.RegisterExtensionsRequest(
            extensions=[ext_descriptor]
        ))
        logger.info(f"Расширение {ext_descriptor.extension} зарегистрировано в perxis")

    @aiocron.crontab('* * * * *', start=False)
    async def register_extension():
        await _register_extension()

    account_channel = grpc.aio.insecure_channel(account_host, interceptors=[interceptor])
    files_channel = grpc.aio.insecure_channel(files_host, interceptors=[interceptor])
    images_channel = grpc.aio.insecure_channel(images_host, interceptors=[interceptor])

    async with grpc.aio.insecure_channel(ext_manager_host, interceptors=[interceptor]) as extensions_manager_channel:
        ext_manager_stub = manager_service_pb2_grpc.ExtensionManagerServiceStub(extensions_manager_channel)

        registered_extensions: manager_service_pb2.ListRegisteredExtensionsResponse = await ext_manager_stub.ListRegisteredExtensions(
           manager_service_pb2.ListRegisteredExtensionsRequest()
        )

        for ext in registered_extensions.extensions:
            if ext.extension == ext_descriptor.extension:
                if ext.version != ext_descriptor.version:
                    await ext_manager_stub.UnregisterExtensions(
                        manager_service_pb2.UnregisterExtensionsRequest(
                            extensions=[ext_descriptor]
                        )
                    )

                    await _register_extension()

                    logger.info(
                        f"Изменилась версия расширения {ext_descriptor.extension} ({ext.version} "
                        f"-> {ext_descriptor.version})"
                    )

        register_extension.start()

        async with grpc.aio.insecure_channel(content_host, interceptors=[interceptor]) as content_channel:

            collections_stub = collections_pb2_grpc.CollectionsStub(content_channel)
            roles_stub = roles_pb2_grpc.RolesStub(content_channel)
            clients_stub = clients_pb2_grpc.ClientsStub(content_channel)
            environments_stub = environments_pb2_grpc.EnvironmentsStub(content_channel)
            items_stub = items_pb2_grpc.ItemsStub(content_channel)
            spaces_stub = spaces_pb2_grpc.SpacesStub(content_channel)
            files_stub = files_pb2_grpc.FilesStub(files_channel)
            images_stub = images_pb2_grpc.ImagesStub(images_channel)
            references_stub = references_pb2_grpc.ReferencesStub(content_channel)
            users_stub = users_pb2_grpc.UsersStub(account_channel)
            organizations_stub = organizations_pb2_grpc.OrganizationsStub(account_channel)
            members_stub = members_pb2_grpc.MembersStub(account_channel)
            locales_stub = locales_pb2_grpc.LocalesStub(content_channel)
            invitations_stub = invitations_pb2_grpc.InvitationsStub(content_channel)
            collaborators_stub = collaborators_pb2_grpc.CollaboratorsStub(content_channel)

            server = grpc.aio.server(futures.ThreadPoolExecutor(max_workers=10))

            servicer = servicer_cls(
                ext_manager_service=ext_manager_stub,
                collections_service=collections_stub,
                environments_service=environments_stub,
                roles_service=roles_stub,
                clients_service=clients_stub,
                items_service=items_stub,
                spaces_service=spaces_stub,
                files_service=files_stub,
                images_service=images_stub,
                references_service=references_stub,
                users_service=users_stub,
                organizations_service=organizations_stub,
                members_service=members_stub,
                locales_service=locales_stub,
                invitations_service=invitations_stub,
                collaborators_service=collaborators_stub,
                channel=content_channel,
            )

            extension_service_pb2_grpc.add_ExtensionServiceServicer_to_server(servicer, server)
            operation_service_pb2_grpc.add_OperationServiceServicer_to_server(servicer, server)

            server.add_insecure_port("[::]:50051")
            await server.start()
            await server.wait_for_termination()

            await account_channel.close()
            await files_channel.close()
            await images_channel.close()


def bootstrap(
        ext_descriptor: manager_service_pb2.ExtensionDescriptor,
        servicer_cls: extension_service_pb2_grpc.ExtensionServiceServicer,
        ext_manager_host: str = "extension-manager:9030",
        content_host: str = "content:9020",
        account_host: str = "account:9010",
        files_host: str = "files:8003",
        images_host: str = "images:8005"
):
    """
        Функция для инициализации и запуска сервиса расширения.

        Пример использования в расширении:
        ```
        def main():
            ext_descriptor = utils.get_extension_descriptor(
                ext_host="demo-ext-backend:50051",
                ext_id=extension.ID,
                ext_name=extension.NAME,
                ext_version=extension.VERSION,
                ext_description=extension.DESCRIPTION,
                ext_version_description=extension.VERSION_DESCRIPTION
            )

            bootstrap(ext_descriptor, Servicer, ext_manager_host, content_host)
        ```

        Arguments:
              ext_descriptor (manager_service_pb2.ExtensionDescriptor): Дескриптор расширения
              servicer_cls (extension_service_pb2_grpc.ExtensionServiceServicer): класс (не экземпляр!) сервиса
              ext_manager_host (str): адрес сервиса менеджера расширений
              content_host (str): адрес сервиса контента
              account_host (str): адрес сервиса аккаунтов
              files_host (str): адрес сервиса файлов
              images_host (str): адрес сервиса работы с изображениями
    """
    logger.info(f"Инициализация сервиса расширения {ext_descriptor.extension}")

    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(
            _main(
                ext_descriptor=ext_descriptor,
                servicer_cls=servicer_cls,
                ext_manager_host=ext_manager_host,
                content_host=content_host,
                account_host=account_host,
                files_host=files_host,
                images_host=images_host,
            )
        )
    finally:
        loop.close()
        loop.stop()
        logging.info('Successfully shutdown service')
