"""
    Модуль содержит базовый класс сервисов расширений - ExtensionService
"""


import grpc
import uuid
import typing
import asyncio
import logging
import aiocron
import datetime
import dataclasses

from google.protobuf import timestamp_pb2, any_pb2, wrappers_pb2

from perxis.extensions import extension_service_pb2, extension_service_pb2_grpc, extension_pb2, manager_service_pb2_grpc
from perxis.collaborators import collaborators_pb2_grpc
from perxis.invitations import invitations_pb2_grpc
from perxis.locales import locales_pb2_grpc
from perxis.organizations import organizations_pb2_grpc
from perxis.members import members_pb2_grpc
from perxis.users import users_pb2_grpc
from perxis.references import references_pb2_grpc
from perxis.images import images_pb2_grpc
from perxis.files import files_pb2_grpc
from perxis.roles import roles_pb2_grpc, roles_pb2
from perxis.items import items_pb2_grpc
from perxis.clients import clients_pb2_grpc, clients_pb2
from perxis.spaces import spaces_pb2_grpc
from perxis.common import operation_pb2, operation_service_pb2_grpc, operation_service_pb2, error_pb2
from perxis.collections import collections_pb2_grpc, collections_pb2
from perxis.environments import environments_pb2_grpc
from perxis.extensions.extension_setup import ExtensionSetup
from perxis.extensions.item_models import AbstractItem


def generate_operation_id() -> str:
    """
        Функция для генерации идентификатора операции

        Returns:
            str
    """

    return str(uuid.uuid4())


def datetime_to_timestamp(dt: datetime) -> timestamp_pb2.Timestamp:
    """
        Функция для преобразования объекта datetime в объект timestamp_pb2.Timestamp

        Arguments:
            dt (datetime)
        Returns:
            timestamp_pb2.Timestamp
    """

    timestamp = dt.timestamp()
    seconds = int(timestamp)
    nanos = int(timestamp % 1 * 1e9)

    return timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)


@dataclasses.dataclass
class OperationMeta:
    """
        Класс для хранения метаданных операции

        Attributes:
            task (asyncio.Task): ссылка на task операции
            operation_id (str): идентификатор операции
            description (str): описание операции
            created_at (datetime.datetime): дата создания операции
            modified_at (datetime.datetime): дата изменения метаданных операции
            response (str | None): результат операции
            errors (list[str]): возникшие во время выполнения операции ошибки
            was_finished (bool): флаг завершения операции
            metadata (dict): любые прочие метаданные операции
    """

    task: asyncio.Task

    operation_id: str
    description: str

    created_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)
    modified_at: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)

    response: typing.Optional[str] = None
    errors: typing.Optional[list[str]] = None

    was_finished: bool = False
    metadata: dict[str, typing.Any] = dataclasses.field(default_factory=dict)

    def mark_cancelled(self):
        """
            Функция для отмены операции
        """

        self.was_finished = True
        self.response = "Отменено"

    def mark_finished(self, errors: typing.Optional[list[str]] = None):
        """
            Функция для завершения операции

            Arguments:
                errors (list[str] | None): список ошибок
        """

        self.modified_at = datetime.datetime.now()
        self.was_finished = True
        self.errors = errors
        self.response = None if errors else "OK"

    def to_operation(self):
        """
            Функция для преобразования объекта операции в формат perxis

            Returns:
                operation_pb2.Operation
        """

        packed_response = any_pb2.Any()
        packed_response.Pack(wrappers_pb2.StringValue(value=self.response or "PENDING"))

        return operation_pb2.Operation(
            id=self.operation_id,
            description=self.description,
            created_at=datetime_to_timestamp(self.created_at),
            modified_at=datetime_to_timestamp(self.modified_at),
            done=self.was_finished,
            metadata=self.metadata,
            response=packed_response,
            error=error_pb2.Error(
                message="; ".join(self.errors)
            ) if self.errors else None
        )


class ExtensionService(
    extension_service_pb2_grpc.ExtensionServiceServicer, operation_service_pb2_grpc.OperationServiceServicer
):
    """
        Базовый класс для сервисов расширений

        Attributes:
            extension_id (str): идентификатор расширения
            collections (list[collections_pb2.Collection]): список коллекций расширения
            roles (list[roles_pb2.Role]): список ролей расширения
            clients (list[clients_pb2.Client]): список клиентов расширения
            actions (list[dict]): список действий расширения
            items (list[AbstractItem]): список управляемых расширением item'ов
            __operations (dict[str, OperationMeta]): маппинг с данными операций
            logger (logging.Logger): логгер расширения
            ext_manager_service (manager_service_pb2_grpc.ExtensionManagerServiceStub): ссылка на сервис менеджера расширений
            collections_service (collections_pb2_grpc.CollectionsStub): ссылка на сервис коллекций
            environments_service (environments_pb2_grpc.EnvironmentsStub): ссылка на сервис окружений
            roles_service (roles_pb2_grpc.RolesStub): ссылка на сервис ролей
            clients_service (clients_pb2_grpc.ClientsStub): ссылка на сервис клиентов
            items_service (items_pb2_grpc.ItemsStub): ссылка на сервис item'ов
            spaces_service (spaces_pb2_grpc.SpacesStub): ссылка на сервис пространств
            files_service (files_pb2_grpc.FilesStub): ссылка на сервис файлов
            images_service (images_pb2_grpc.ImagesStub): ссылка на сервис изображений
            references_service (references_pb2_grpc.ReferencesStub): ссылка на сервис для работы с объектами Reference
            users_service (users_pb2_grpc.UsersStub): ссылка на сервис пользователей
            organizations_service (organizations_pb2_grpc.OrganizationsStub): ссылка на сервис организаций
            members_service (members_pb2_grpc.MembersStub): ссылка на сервис работы с пользователями организаций
            locales_service (locales_pb2_grpc.LocalesStub): ссылка на сервис локалей и переводов
            invitations_service (invitations_pb2_grpc.InvitationsStub): ссылка на сервис приглашений
            collaborators_service (collaborators_pb2_grpc.CollaboratorsStub): ссылка на сервис коллабораторов
            channel (grpc.Channel): ссылка на grpc канал
            extension_setup (ExtensionSetup): объект класса ExtensionSetup для управления данными расширения
    """

    extension_id: str
    collections: list[collections_pb2.Collection] = []
    roles: list[roles_pb2.Role] = []
    clients: list[clients_pb2.Client] = []
    actions: list[dict] = []
    items: list[AbstractItem] = []

    __operations: dict[str, OperationMeta]

    def __init__(self,
                 ext_manager_service: manager_service_pb2_grpc.ExtensionManagerServiceStub,
                 collections_service: collections_pb2_grpc.CollectionsStub,
                 environments_service: environments_pb2_grpc.EnvironmentsStub,
                 roles_service: roles_pb2_grpc.RolesStub,
                 clients_service: clients_pb2_grpc.ClientsStub,
                 items_service: items_pb2_grpc.ItemsStub,
                 spaces_service: spaces_pb2_grpc.SpacesStub,
                 files_service: files_pb2_grpc.FilesStub,
                 images_service: images_pb2_grpc.ImagesStub,
                 references_service: references_pb2_grpc.ReferencesStub,
                 users_service: users_pb2_grpc.UsersStub,
                 organizations_service: organizations_pb2_grpc.OrganizationsStub,
                 members_service: members_pb2_grpc.MembersStub,
                 locales_service: locales_pb2_grpc.LocalesStub,
                 invitations_service: invitations_pb2_grpc.InvitationsStub,
                 collaborators_service: collaborators_pb2_grpc.CollaboratorsStub,
                 channel: grpc.Channel,
    ):
        self.logger = logging.getLogger(__name__)
        self.ext_manager_service = ext_manager_service
        self.collections_service = collections_service
        self.environments_service = environments_service
        self.roles_service = roles_service
        self.clients_service = clients_service
        self.items_service = items_service
        self.spaces_service = spaces_service
        self.files_service = files_service
        self.images_service = images_service
        self.references_service = references_service
        self.users_service = users_service
        self.organizations_service = organizations_service
        self.members_service = members_service
        self.locales_service = locales_service
        self.invitations_service = invitations_service
        self.collaborators_service = collaborators_service
        self.channel = channel

        self.extension_setup = ExtensionSetup(
            self.collections_service, self.environments_service,
            self.roles_service, self.clients_service, self.items_service,
        )

        self.__operations = {}

        for collection in self.collections or []:
            self.extension_setup.add_collection(collection)

        for role in self.roles or []:
            self.extension_setup.add_role(role)

        for client in self.clients or []:
            self.extension_setup.add_client(client)

        for action in self.actions or []:
            self.extension_setup.add_action(action)

        services_list = [
            (service_name, getattr(self, service_name))
            for service_name
            in self.__dict__
            if service_name.endswith("_service")
        ]
        services_list.append(("channel", self.channel, ))

        for item in self.items:
            for rule in item.rules:
                rule.bind_services(services_list)

        self.extension_setup.set_items(self.items)

        @aiocron.crontab('0 * * * *', start=True)
        async def remove_old_operations():
            self.remove_old_operations()

    @property
    def operations(self):
        return self.__operations

    def remove_old_operations(self):
        """
            Функция для удаления stale операций. Запускается раз в час. Удаляет все операции
            которые были завершены больше часа назад
        """

        self.logger.info("Удаление старых операций")

        ids = []
        now = datetime.datetime.now()

        for operation_meta in self.__operations.values():
            task_is_not_running = operation_meta.task.done() or operation_meta.task.cancelled()

            # Если task фактически не работает то операцию нужно пометить как выполненную. Это может произойти в случае
            # неотловленного исключения, например
            if task_is_not_running:
                operation_meta.was_finished = True

            if not operation_meta.was_finished:
                continue

            time_delta = now - operation_meta.modified_at

            # Для операций которые завершены больше часа назад
            if time_delta.seconds < 60 * 60:
                continue

            # Если по какой-то причине операция была помечена как завершенная но task ещё активен
            need_to_cancel_task = not any(
                (operation_meta.task.done(), operation_meta.task.cancelled())
            )

            if need_to_cancel_task:
                try:
                    operation_meta.task.cancel()
                except Exception as e:
                    self.logger.error("Не удалось отменить task")
                    self.logger.exception(e)

                    continue

            ids.append(operation_meta.operation_id)

            self.logger.info(f"Операция {operation_meta.operation_id} помечена на удаление")

        for operation_id in ids:
            del self.__operations[operation_id]

        self.logger.info("Удаление старых операций завершено")

    def result_log(self, operation, operation_id: str, request, errors: list[str]):
        """
            Вспомогательная функция для логгирования резултата выполнения операции

            Arguments:
                  operation (str): название операции
                  operation_id (str): идентификатор операции
                  request: объект запроса
                  errors (list[str]): ошибки выполнения операции
        """

        log_func = self.logger.error if errors else self.logger.info

        log_func(
            "Операция %s (%s) расширения %s для окружения %s пространства %s завершена %s"
            % (
                operation,
                operation_id,
                self.extension_id,
                request.env_id,
                request.space_id,
                "с ошибками: " + "; ".join(errors) if errors else "успешно"
            )
        )

    def ext_request_results_from_exception(self, e: Exception):
        """
            Вспомогательная функция для преобразования объектов исключений в объект результата запроса

            Arguments:
                e (Exception): любое исключение
            Returns:
                extension_service_pb2.ExtensionRequestResult
        """

        return [
            extension_service_pb2.ExtensionRequestResult(
                extension=self.extension_id,
                state=extension_service_pb2.ExtensionRequestResult.State.ERROR,
                error=str(e),
                msg=None
            )
        ]

    def get_operation_meta(self, operation_id: str) -> typing.Optional[OperationMeta]:
        """
            Получение объекта операции по идентификатору

            Arguments:
                    operation_id (str): идентификатор операции

            Returns:
                   OperationMeta | None
        """

        return self.__operations.get(operation_id)

    def set_operation_meta(self, operation_id: str, description: str, task: asyncio.Task):
        """
            Функция для сохранения данных об операции
            Arguments:
                  operation_id (str): идентификатор операции
                  description (str): описание операции
                  task (asyncio.Task): объект task операции
        """

        self.__operations[operation_id] = OperationMeta(
            operation_id=operation_id,
            task=task,
            description=description
        )

    def mark_operation_as_finished(self, operation_id: str, errors: list[str] | None = None):
        """
            Функция для завершения операции

            Arguments:
                operation_id (str): идентификатор операции
                errors (list[str] | None): ошибки выполнения операции
        """

        if operation_id in self.operations:
            self.operations[operation_id].mark_finished(errors)
        else:
            self.logger.error(f"Операция {operation_id} не найдена!")

    async def _Install(self, operation_id: str, request: extension_service_pb2.InstallRequest, context):
        """
            Действия метода Install сервиса расширений. Вызывается после всех необходимых действий с
            формированием данных об операции. В случае необходимости можно переопределить

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.InstallRequest): объект запроса
                   context: контекст
        """

        errors_list = await self.extension_setup.install(
            request.space_id, request.env_id, request.force
        )

        errors_list += await self.additional_install_operations(operation_id, request, context)

        self.result_log("установки", operation_id, request, errors_list)

        self.mark_operation_as_finished(operation_id, errors_list)

    async def additional_install_operations(self, operation_id: str, request: extension_service_pb2.InstallRequest, context) -> list[str]:
        """
            Функция для возможности добавления любой дополнительной логики после завершения всех стандартных
            процедур установки. Можно переопределять. Возвращает массив с ошибками.

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.InstallRequest): объект запроса
                   context: контекст

            Returns:
                list[str]
        """

        return []

    async def Install(self, request: extension_service_pb2.InstallRequest, context):
        """
            Реализация метода Install класса extension_service_pb2_grpc.ExtensionServiceServicer. Содержит
            в себе формирование данных об операции. Переопределять не следует, для изменения логики Install
            лучше работать с методами _Install если нужно всё радикально изменить или с методом
            additional_install_operations для добавления чего-то дополнительного
        """

        operation_id = generate_operation_id()
        operation_description = "Установка расширения %s для окружения %s пространства %s. %s force" % (
            self.extension_id,
            request.env_id,
            request.space_id,
            "С" if request.force else "Без"
        )

        self.logger.info(operation_description)

        install_task = asyncio.create_task(self._Install(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, install_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _Uninstall(self, operation_id: str, request: extension_service_pb2.UninstallRequest, context):
        """
            Действия метода _Uninstall сервиса расширений. Вызывается после всех необходимых действий с
            формированием данных об операции. В случае необходимости можно переопределить

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.UninstallRequest): объект запроса
                   context: контекст
        """

        errors_list: list[str] = await self.extension_setup.uninstall(request.space_id, request.env_id, request.remove)
        errors_list += await self.additional_uninstall_operations(operation_id, request, context)

        self.result_log("удаления", operation_id, request, errors_list)

        self.mark_operation_as_finished(operation_id, errors_list)

    async def additional_uninstall_operations(self, operation_id: str, request: extension_service_pb2.UninstallRequest, context) -> list[str]:
        """
            Функция для возможности добавления любой дополнительной логики после завершения всех стандартных
            процедур удаления. Можно переопределять. Возвращает массив с ошибками.

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.UninstallRequest): объект запроса
                   context: контекст

            Returns:
                list[str]
        """

        return []

    async def Uninstall(self, request: extension_service_pb2.UninstallRequest, context):
        """
            Реализация метода Uninstall класса extension_service_pb2_grpc.ExtensionServiceServicer. Содержит
            в себе формирование данных об операции. Переопределять не следует, для изменения логики Uninstall
            лучше работать с методами _Uninstall если нужно всё радикально изменить или с методом
            additional_uninstall_operations для добавления чего-то дополнительного
        """

        operation_id = generate_operation_id()
        operation_description = "Удаление расширения %s для окружения %s пространства %s. %s remove" % (
            self.extension_id,
            request.env_id,
            request.space_id,
            "С" if request.remove else "Без"
        )

        self.logger.info(operation_description)

        uninstall_task = asyncio.create_task(self._Uninstall(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, uninstall_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _Check(self, operation_id: str, request: extension_service_pb2.CheckRequest, context):
        """
            Действия метода _Check сервиса расширений. Вызывается после всех необходимых действий с
            формированием данных об операции. В случае необходимости можно переопределить

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.CheckRequest): объект запроса
                   context: контекст
        """

        errors_list: list[str] = await self.extension_setup.check(request.space_id, request.env_id)

        self.result_log("проверки", operation_id, request, errors_list)

        self.__operations[operation_id].mark_finished(errors_list)

    async def additional_check_operations(self, operation_id: str, request: extension_service_pb2.CheckRequest, context) -> list[str]:
        """
            Функция для возможности добавления любой дополнительной логики после завершения всех стандартных
            процедур проверки. Можно переопределять. Возвращает массив с ошибками.

            Arguments:
                   operation_id (str): идентификатор операции
                   request (extension_service_pb2.CheckRequest): объект запроса
                   context: контекст

            Returns:
                list[str]
        """

        return []

    async def Check(self, request: extension_service_pb2.CheckRequest, context):
        """
            Реализация метода Check класса extension_service_pb2_grpc.ExtensionServiceServicer. Содержит
            в себе формирование данных об операции. Переопределять не следует, для изменения логики Check
            лучше работать с методами _Check если нужно всё радикально изменить или с методом
            additional_check_operations для добавления чего-то дополнительного
        """

        operation_id = generate_operation_id()
        operation_description = "Проверка расширения %s для окружения %s пространства %s" % (
            self.extension_id,
            request.env_id,
            request.space_id,
        )

        check_task = asyncio.create_task(self._Check(operation_id, request, context))

        self.set_operation_meta(operation_id, operation_description, check_task)

        return self.get_operation_meta(operation_id).to_operation()

    async def _dispatch_action(
            self, request: extension_pb2.ActionRequest, context
    ) -> extension_service_pb2.ActionResponse:
        """
            Метод для получение нужной ф-ции для действия и её вызова

            Arguments:
               request (extension_pb2.ActionRequest): объект запроса
               context: контекст

            Returns:
               extension_service_pb2.ActionResponse
        """

        action_id = request.action.split("/")[-1]

        func_name = f"action_{action_id}"

        if not hasattr(self, func_name):
            response = extension_service_pb2.ActionResponse(
                state=extension_service_pb2.ActionResponse.State.ERROR,
                title="Невозможно выполнить действие",
                error=f"В расширении отсутсвует функция {action_id}"
            )
        else:
            func = getattr(self, func_name)

            response = await func(request, context)

        return response

    async def Action(self, request: extension_pb2.ActionRequest, context):
        """
            Реализация метода Action класса extension_service_pb2_grpc.ExtensionServiceServicer

            Arguments:
                request (extension_pb2.ActionRequest): объект запроса
                context: контекст
        """

        operation_description = "Действие %s для окружения %s пространства %s" % (
            request.action,
            request.env_id,
            request.space_id,
        )

        self.logger.info(operation_description)

        response = await self._dispatch_action(request, context)

        if response.state == extension_service_pb2.ActionResponse.State.ERROR:
            error_description = "Ошибка действия %s в расширении %s: %s" % (
                request.action,
                self.extension_id,
                response.error
            )

            self.logger.error(error_description)

        return response

    async def Get(self, request: operation_service_pb2.GetOperationRequest, context):
        """
            Реализация метода Get класса operation_service_pb2_grpc.OperationServiceServicer. Используется
            для получения данных о выполняемой операции

            Arguments:
                request (operation_service_pb2.GetOperationRequest): объект запроса
                context: контекст
        """

        operations_meta = self.get_operation_meta(request.operation_id)

        if not operations_meta:
            error_description = "Ошибка проверки операции %s в расширении %s - не найдена" % (
                request.operation_id,
                self.extension_id,
            )

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.error(error_description)

            return None

        self.logger.info(
            f"Статус операции {operations_meta.operation_id}:\n"
            f"{operations_meta.description}\n"
            f"{'выполняется' if not operations_meta.was_finished else 'завершена'}, \n"
            f"{'без ошибок' if not operations_meta.errors else 'ошибки: ' + '; '.join(operations_meta.errors)}"
        )

        return operations_meta.to_operation()

    def Cancel(self, request: operation_service_pb2.CancelOperationRequest, context):
        """
            Реализация метода Cancel класса operation_service_pb2_grpc.OperationServiceServicer. Используется
            для отмены выполнения операции

            Arguments:
                request (operation_service_pb2.CancelOperationRequest): объект запроса
                context: контекст
        """

        operations_meta = self.get_operation_meta(request.operation_id)

        if not operations_meta:
            error_description = "Не удалось удалить операцию %s в расширении %s - не найдена" % (
                request.operation_id,
                self.extension_id,
            )

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.error(
                error_description
            )

            return None

        try:
            operations_meta.task.cancel()
            operations_meta.mark_cancelled()
        except Exception as e:
            error_description = "Во время отмены операции %s расширении %s произошла ошибка %s" % (
                request.operation_id,
                self.extension_id,
                str(e)
            )

            self.logger.error(error_description)

            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(error_description)

            self.logger.exception(e)

            return None

        self.logger.info(
            "Операция %s в расширении %s отменена" % (
                request.operation_id,
                self.extension_id,
            )
        )

        return operations_meta.to_operation()
