"""
Модуль содержит вспомогательные функции для сервисов расширений
"""

from typing import Optional

from perxis.extensions import manager_service_pb2
from perxis.extensions.item_models import DataSourceItem, DataSourceData, SyncPolicyItem, SyncPolicyData


def datasource_items_from_collections(
    collections_map: dict[str, str],
    datasource_data: DataSourceData,
    override_for_collections: dict[str, DataSourceData] | None = None,
    with_update: bool = True,
    with_delete: bool = True,
) -> list[DataSourceItem]:
    """
        Создание записей источников данных на базе маппинга коллекций.

        Arguments:
            collections_map: Dict вида [collection_id, collection_name]
            datasource_data: Dict со структурой DataSourceData для item'ов коллекций
            override_for_collections: Dict для переопределения значений для отдельных коллекций
                вида [collection_id, DataSourceData]
            with_update: Флаг для возможности обновления существующих записей расширением. Может быть переопределён
                в override_for_collections
            with_delete: Флаг для возможности удаления существующих записей расширением. Может быть переопределён
                в override_for_collections
        Returns:
            list[DataSourceItem]
    """

    if not isinstance(override_for_collections, dict):
        override_for_collections = {}

    items = []

    for collection_id in collections_map:
        if not collection_id:
            continue

        kwargs = {
            "with_update": with_update,
            "with_delete": with_delete,
            **override_for_collections.get(collection_id, datasource_data),
            "collection_id": collection_id,
        }

        items.append(DataSourceItem(**kwargs))

    return items


def sync_policies_from_collections(
    collections_map: dict[str, str],
    sync_policies_data: SyncPolicyData,
    override_for_collections: dict[str, SyncPolicyData] | None = None,
    with_update: bool = True,
    with_delete: bool = True,
) -> list[SyncPolicyItem]:
    """
        Создание записей синхронизации коллекций на базе маппинга коллекций

        Arguments:
            collections_map: Dict вида [collection_id, collection_name]
            sync_policies_data: Dict со структурой SyncPolicyData для item'ов коллекций
            override_for_collections: Dict для переопределения значений для отдельных коллекций
                вида [collection_id, SyncPolicyData]
            with_update: Флаг для возможности обновления существующих записей расширением. Может быть переопределён
                в override_for_collections
            with_delete: Флаг для возможности удаления существующих записей расширением. Может быть переопределён
                в override_for_collections
        Returns:
            list[SyncPolicyItem]
    """

    if not isinstance(override_for_collections, dict):
        override_for_collections = {}

    items = []

    for collection_id, collection_name in collections_map.items():
        if not collection_id:
            continue

        kwargs = {
            "with_update": with_update,
            "with_delete": with_delete,
            "name": collection_name,
            **override_for_collections.get(collection_id, sync_policies_data),
            "collection_id": collection_id,
        }

        items.append(SyncPolicyItem(**kwargs))

    return items


def get_extension_descriptor(
        ext_host: str, ext_id: str, ext_name: str, ext_description: str,
        ext_version: str, ext_version_description: str,
        ext_deps: Optional[list[str]] = None
) -> manager_service_pb2.ExtensionDescriptor:
    """
        Функция для получения дескриптора расширения на основании его данных

        Arguments:
            ext_host (str): адрес расширения
            ext_id (str): идентификатор расширения
            ext_name (str): название расширения
            ext_version (str): версия расширения
            ext_version_description (str): описание расширения
            ext_deps (list[str] | None): зависимости расширения

        Returns:
            manager_service_pb2.ExtensionDescriptor
    """

    if not ext_deps:
        ext_deps = []

    return manager_service_pb2.ExtensionDescriptor(
        extension=ext_id,
        title=ext_name,
        description=ext_description,
        version=ext_version,
        version_description=ext_version_description,
        deps=ext_deps,
        url=ext_host
    )
